/*
 * *********************************************************
 * Copyright (c) 2009 - 2010, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: 20.10.2010
 * Author(s):
 * Gero
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai;

import java.util.concurrent.BlockingDeque;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.TimeUnit;

import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.moduli.exceptions.InitModuleException;
import edu.dhbw.mannheim.tigers.moduli.exceptions.ModuleNotFoundException;
import edu.dhbw.mannheim.tigers.moduli.exceptions.StartModuleException;
import edu.dhbw.mannheim.tigers.sumatra.model.SumatraModel;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.AIInfoFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.BaseAiFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.MetisAiFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.SimpleWorldFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.WorldFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.ETeamColor;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.TacticalField;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.referee.RefereeMsg;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.Metis;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.timer.ETimable;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.timer.SumatraTimer;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.oextkal.Director;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.observer.IAIObserver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AAgent;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.ACam;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AReferee;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.ATimer;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AWorldPredictor;
import edu.dhbw.mannheim.tigers.sumatra.util.FpsCounter;


/**
 * This is the one-and-only agent implementation, which controls the AI-sub-modules, and sends out our MechWarriors in
 * the endless battle for fame and glory!
 * 
 * @author Gero
 */
public class Agent extends AAgent implements Runnable
{
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Logger
	private static final Logger					log					= Logger.getLogger(Agent.class.getName());
	
	private static final int						QUEUE_LENGTH		= 1;
	
	private static final long						WF_TIMEOUT			= 1000;
	
	// Source
	private final SumatraModel						model					= SumatraModel.getInstance();
	private SumatraTimer								timer					= null;
	private ACam										cam					= null;
	
	private final BlockingDeque<WorldFrame>	freshWorldFrames	= new LinkedBlockingDeque<WorldFrame>(QUEUE_LENGTH);
	
	
	// AI
	private Thread										nathan;
	
	private AReferee									referee				= null;
	
	private AIInfoFrame								previousAIFrame	= null;
	private RefereeMsg								latestRefereeMsg	= null;
	
	/** {@link Metis} */
	private Metis										metis;
	
	private FpsCounter								fpsCounter			= new FpsCounter();
	
	private ETimable									timableAgent;
	private ETimable									timableMetis;
	private ETimable									timableNotify;
	
	
	private long										lastRefMsgCounter	= -1;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @param subnodeConfiguration
	 */
	public Agent(final SubnodeConfiguration subnodeConfiguration)
	{
	}
	
	
	// --------------------------------------------------------------------------
	// --- lifecycle ------------------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void initModule() throws InitModuleException
	{
		if (MODULE_ID.equals(getId()))
		{
			timableAgent = ETimable.AGENT_Y;
			timableMetis = ETimable.METIS_Y;
			timableNotify = ETimable.AI_NOTIFY_Y;
			setTeamColor(ETeamColor.YELLOW);
		}
		
		try
		{
			cam = (ACam) model.getModule(ACam.MODULE_ID);
			
			AWorldPredictor predictor = (AWorldPredictor) model.getModule(AWorldPredictor.MODULE_ID);
			predictor.addWorldFrameConsumer(this);
			
			referee = (AReferee) model.getModule(AReferee.MODULE_ID);
			referee.addRefereeMsgConsumer(this);
		} catch (final ModuleNotFoundException err)
		{
			log.error("Unable to find one or more modules!");
		}
	}
	
	
	@Override
	public void startModule() throws StartModuleException
	{
		lastRefMsgCounter = -1;
		try
		{
			timer = ((SumatraTimer) model.getModule(ATimer.MODULE_ID));
		} catch (final ModuleNotFoundException err)
		{
			log.debug("No timer found.");
		}
		
		metis = new Metis();
		
		nathan = new Thread(this, "AI_Nathan_" + getId());
		nathan.start();
		log.trace("Nathan started");
	}
	
	
	/**
	 * @param msg The recently received message
	 * @return Whether this message does really new game-state information
	 * @author FriederB
	 */
	private boolean isNewMessage(final RefereeMsg msg)
	{
		if (msg.getCommandCounter() != lastRefMsgCounter)
		{
			lastRefMsgCounter = msg.getCommandCounter();
			return true;
		}
		
		return false;
	}
	
	
	/**
	 *
	 */
	@Override
	public void run()
	{
		long id = 0;
		while (!Thread.currentThread().isInterrupted())
		{
			// ### Get latest worldframe
			WorldFrame wf;
			try
			{
				wf = freshWorldFrames.pollLast(WF_TIMEOUT, TimeUnit.MILLISECONDS);
				
				if ((wf == null))
				{
					continue;
				}
			} catch (final InterruptedException err)
			{
				// No error here...
				break;
			}
			
			id = wf.getId().getFrameNumber();
			
			timer.start(timableAgent, id);
			
			// ### Take the first of the referee-messages
			RefereeMsg refereeMsg = latestRefereeMsg;
			
			RefereeMsg newRefereeMsg = null;
			if ((refereeMsg != null) && isNewMessage(refereeMsg))
			{
				newRefereeMsg = refereeMsg;
			}
			
			BaseAiFrame baseAiFrame = new BaseAiFrame(wf, newRefereeMsg, refereeMsg, previousAIFrame, getTeamColor());
			
			if (previousAIFrame == null)
			{
				// Skip first frame
				previousAIFrame = generateAIInfoFrame(baseAiFrame);
				continue;
			}
			
			previousAIFrame.cleanUp();
			
			// ### Process!
			try
			{
				// Analyze
				timer.start(timableMetis, id);
				MetisAiFrame metisAiFrame = metis.process(baseAiFrame);
				timer.stop(timableMetis, id);
				
				// Choose and calculate behavior
				
				timer.start(timableNotify, id);
				// ### Populate used AIInfoFrame (for visualization etc)
				AIInfoFrame frame = new AIInfoFrame(metisAiFrame, fpsCounter.getAvgFps());
				fpsCounter.newFrame();
				
				notifyNewAIInfoFrame(frame);
				previousAIFrame = frame;
				timer.stop(timableNotify, id);
			} catch (final Exception ex)
			{
				// # Notify observers (gui) about errors...
				notifyNewAIException(ex, generateAIInfoFrame(baseAiFrame), previousAIFrame);
				
				// # Undo everything we've done this cycle to restore previous state
				// - RefereeMsg
				if (refereeMsg != null)
				{
					lastRefMsgCounter--;
				}
				
				// # Prepare next cycle
				timer.stop(timableAgent, id);
				
				previousAIFrame.cleanUp();
				continue;
			}
			
			
			// ### End cycle
			timer.stop(timableAgent, id);
		}
	}
	
	
	private AIInfoFrame generateAIInfoFrame(final BaseAiFrame baseAiFrame)
	{
		return new AIInfoFrame(new MetisAiFrame(baseAiFrame, new TacticalField(
				baseAiFrame.getWorldFrame())), 0);
	}
	
	
	@Override
	public void stopModule()
	{
		nathan.interrupt();
	}
	
	
	@Override
	public void deinitModule()
	{
		if (referee != null)
		{
			referee = null;
		}
		
		if (cam != null)
		{
			cam = null;
		}
		
		previousAIFrame = null;
	}
	
	
	// --------------------------------------------------------------------------
	// --- observer methods -----------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void onNewRefereeMsg(final RefereeMsg msg)
	{
		latestRefereeMsg = msg;
	}
	
	
	@Override
	public synchronized void onNewSimpleWorldFrame(final SimpleWorldFrame worldFrame)
	{
	}
	
	
	@Override
	public void onNewWorldFrame(final WorldFrame wFrame)
	{
		if (wFrame.getTeamColor() == getTeamColor())
		{
			freshWorldFrames.pollLast();
			freshWorldFrames.addFirst(wFrame);
		}
	}
	
	
	@Override
	public void onVisionSignalLost(final SimpleWorldFrame emptyWf)
	{
		WorldFrame wf = Director.createWorldFrame(emptyWf, getTeamColor());
		BaseAiFrame bFrame = new BaseAiFrame(wf, null, latestRefereeMsg, previousAIFrame, getTeamColor());
		// ### Populate used AIInfoFrame (for visualization etc)
		notifyNewAIInfoFrame(generateAIInfoFrame(bFrame));
	}
	
	
	@Override
	public void onStop()
	{
		nathan.interrupt();
	}
	
	
	// --------------------------------------------------------------------------
	// --- AI Visualization -----------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * This function is used to notify the last {@link AIInfoFrame} to visualization observers.
	 * 
	 * @param lastAIInfoframe
	 */
	private void notifyNewAIInfoFrame(final AIInfoFrame lastAIInfoframe)
	{
		synchronized (getObservers())
		{
			final AIInfoFrame newAIInfoFrame = new AIInfoFrame(lastAIInfoframe);
			for (final IAIObserver o : getObservers())
			{
				o.onNewAIInfoFrame(newAIInfoFrame);
			}
		}
	}
	
	
	private void notifyNewAIException(final Exception ex, final AIInfoFrame frame, final AIInfoFrame prevFrame)
	{
		synchronized (getObservers())
		{
			AIInfoFrame consFrame = frame;
			if (consFrame != null)
			{
				consFrame = new AIInfoFrame(consFrame);
			}
			AIInfoFrame consPrevFrame = prevFrame;
			if (prevFrame != null)
			{
				consPrevFrame = new AIInfoFrame(consPrevFrame);
			}
			
			for (final IAIObserver o : getObservers())
			{
				o.onAIException(ex, consFrame, consPrevFrame);
			}
		}
	}
	
	
	/**
	 * @return the metis
	 */
	public final Metis getMetis()
	{
		return metis;
	}
}
