/*
 * *********************************************************
 * Copyright (c) 2009 - 2010, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: 03.08.2010
 * Author(s): Malte
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.configuration.HierarchicalConfiguration;

import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.config.AConfigClient;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.oextkal.WorldFramePacker;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AAgent;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.IConfigClient;
import edu.dhbw.mannheim.tigers.sumatra.util.IInstanceableEnum;
import edu.dhbw.mannheim.tigers.sumatra.util.config.ConfigAnnotationProcessor;


/**
 * This holds some static variables to parameterize the AI
 * hard choices - null == usual procedures in classes
 * 
 * @author Oliver Steinbrecher <OST1988@aol.com>, Malte
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public final class AIConfig
{
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	
	// Geometry
	private final IConfigClient	geomClient	= new GeometryConfigClient();
	private volatile Geometry		geometry;
	
	private final AiConfigClient	miscClient;
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	private static class AIConfigHolder
	{
		private static final AIConfig	CONFIG	= new AIConfig();
	}
	
	
	private AIConfig()
	{
		Set<Class<?>> miscClasses = new LinkedHashSet<Class<?>>();
		miscClasses.add(WorldFramePacker.class);
		miscClient = new AiConfigClient("misc", miscClasses);
	}
	
	
	/**
	 * @return
	 */
	public static AIConfig getInstance()
	{
		return AIConfigHolder.CONFIG;
	}
	
	
	/**
	 * @return
	 */
	public static List<IConfigClient> getConfigClients()
	{
		List<IConfigClient> clients = new ArrayList<IConfigClient>();
		clients.add(getInstance().geomClient);
		clients.add(getInstance().miscClient);
		return clients;
	}
	
	
	// --------------------------------------------------------------------------
	// --- IConfigClients -------------------------------------------------------
	// --------------------------------------------------------------------------
	
	
	private final class GeometryConfigClient extends AConfigClient
	{
		private GeometryConfigClient()
		{
			super("geometry", AAgent.GEOMETRY_CONFIG_PATH, AAgent.KEY_GEOMETRY_CONFIG, AAgent.VALUE_GEOMETRY_CONFIG, false);
		}
		
		
		@Override
		public void onLoad(final HierarchicalConfiguration config)
		{
			geometry = new Geometry(config);
		}
	}
	
	
	/**
	 * This general ai config client serves as creator for configs based on {@link ConfigAnnotationProcessor}
	 * 
	 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
	 */
	public static class AiConfigClient extends AConfigClient
	{
		private final ConfigAnnotationProcessor	cap	= new ConfigAnnotationProcessor();
		private final String								name;
		private final Set<Class<?>>					classes;
		
		
		private AiConfigClient(final String name, final Set<Class<?>> classes)
		{
			super(name, AAgent.AI_CONFIG_PATH, AIConfig.class.getName() + "." + name, name + ".xml", true);
			this.name = name;
			this.classes = classes;
		}
		
		
		@Override
		public void onLoad(final HierarchicalConfiguration newConfig)
		{
			cap.loadConfiguration(newConfig);
			cap.apply();
			applySpezis();
		}
		
		
		@Override
		public HierarchicalConfiguration getDefaultConfig()
		{
			return cap.getDefaultConfig(classes, name);
		}
		
		
		/**
		 * Apply all config values with given spezi. If spezi=="", apply default values.
		 * 
		 * @param obj The instance where all fields should be set.
		 * @param spezi
		 */
		public void applyConfigToObject(final Object obj, final String spezi)
		{
			cap.apply(obj, spezi);
		}
		
		
		/**
		 * Apply all know spezis
		 */
		public void applySpezis()
		{
		}
	}
	
	
	@SuppressWarnings("unused")
	private Set<Class<?>> getAllClasses(final Class<? extends Enum<? extends IInstanceableEnum>> enumClazz)
	{
		Set<Class<?>> classes = new LinkedHashSet<Class<?>>();
		Enum<? extends IInstanceableEnum> values[] = enumClazz.getEnumConstants();
		IInstanceableEnum valInstancable[] = (IInstanceableEnum[]) values;
		for (IInstanceableEnum en : valInstancable)
		{
			classes.add(en.getInstanceableClass().getImpl());
		}
		Set<Class<?>> superClasses = new LinkedHashSet<Class<?>>();
		for (Class<?> clazz : classes)
		{
			if (clazz.getSuperclass() != null)
			{
				superClasses.add(clazz.getSuperclass());
			}
		}
		classes.addAll(superClasses);
		return classes;
	}
	
	
	// --------------------------------------------------------------------------
	// --- accessors ------------------------------------------------------------
	// --------------------------------------------------------------------------
	
	
	/**
	 * @return geometry values
	 */
	public static Geometry getGeometry()
	{
		if (AIConfig.getInstance().geometry == null)
		{
			throw new IllegalStateException("geometry is null!");
		}
		return AIConfig.getInstance().geometry;
	}
	
}
