/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Jul 14, 2014
 * Author(s): Lukas Magel
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.autoreferee;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.moduli.AModule;
import edu.dhbw.mannheim.tigers.moduli.exceptions.InitModuleException;
import edu.dhbw.mannheim.tigers.moduli.exceptions.ModuleNotFoundException;
import edu.dhbw.mannheim.tigers.moduli.exceptions.StartModuleException;
import edu.dhbw.mannheim.tigers.sumatra.model.SumatraModel;
import edu.dhbw.mannheim.tigers.sumatra.model.data.airecord.IRecordFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.AIInfoFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.ETeamColor;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.observer.IAIObserver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AAgent;


/**
 * AutoReferee module
 * 
 * @author Lukas Magel
 */
public class AutoReferee extends AModule implements IAIObserver
{
	
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final String	REF_STRINGS_KEY	= "RefereeMsgStrings";
	private static final String	TEAM_AT_FAULT_KEY	= "TEAM_AT_FAULT";
	
	private List<IRefereeCase>		refereeCases		= null;
	private Logger						log					= Logger.getLogger(AutoReferee.class);
	Properties							refMsgs				= null;
	private String						refMsgsPath			= "";
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @param subconfig
	 */
	public AutoReferee(final SubnodeConfiguration subconfig)
	{
		refMsgs = new Properties();
		refMsgsPath = (String) subconfig.getProperty(REF_STRINGS_KEY);
		
		refereeCases = new ArrayList<IRefereeCase>();
		refereeCases.add(new OutOfBoundsCase());
		refereeCases.add(new BallSpeedingCase());
		refereeCases.add(new MultipleDefendersCase(ETeamColor.BLUE));
		refereeCases.add(new MultipleDefendersCase(ETeamColor.YELLOW));
	}
	
	
	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	
	@Override
	public void initModule() throws InitModuleException
	{
		InputStream inStream = null;
		try
		{
			URL msgsURL = AutoReferee.class.getResource(refMsgsPath);
			inStream = msgsURL.openStream();
			refMsgs.load(inStream);
		} catch (IOException err)
		{
			log.error(err.getMessage(), err);
			throw new InitModuleException(err.getMessage(), err);
		} finally
		{
			if (inStream != null)
			{
				try
				{
					inStream.close();
				} catch (IOException err)
				{
					log.warn(err.getMessage(), err);
				}
			}
		}
		
	}
	
	
	@Override
	public void deinitModule()
	{
		
	}
	
	
	@Override
	public void startModule() throws StartModuleException
	{
		for (IRefereeCase refCase : refereeCases)
		{
			refCase.reset();
		}
		
		try
		{
			// register auto referee for new ai frames
			// only use yellow agent, because autoReferee does not use team specific data from ai frames
			AAgent agent = (AAgent) SumatraModel.getInstance().getModule(AAgent.MODULE_ID);
			agent.addObserver(this);
		} catch (ModuleNotFoundException err)
		{
			throw new StartModuleException(err.getMessage(), err);
		}
	}
	
	
	@Override
	public void stopModule()
	{
		
	}
	
	
	@Override
	public void onNewAIInfoFrame(final AIInfoFrame lastAIInfoframe)
	{
		RefereeCaseMsg msg = null;
		for (IRefereeCase refCase : refereeCases)
		{
			msg = refCase.onNewAIInfoFrame(lastAIInfoframe);
			if (msg != null)
			{
				log.warn(buildLogString(msg));
			}
		}
	}
	
	
	private String buildLogString(final RefereeCaseMsg msg)
	{
		StringBuilder stringMsg = new StringBuilder();
		stringMsg.append(refMsgs.get(msg.getMsgType().toString()));
		stringMsg.append(System.lineSeparator());
		stringMsg.append(refMsgs.get(TEAM_AT_FAULT_KEY));
		stringMsg.append(msg.getTeamAtFault().toString());
		return stringMsg.toString();
	}
	
	
	@Override
	public void onAIException(final Exception ex, final IRecordFrame frame, final IRecordFrame prevFrame)
	{
		
	}
	
	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	
}
