/*
 * *********************************************************
 * Copyright (c) 2009 - 2010, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: 10.08.2010
 * Author(s): Bernhard
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.presenter.visualizer;

import java.awt.Component;
import java.awt.event.MouseEvent;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import javax.swing.JCheckBox;
import javax.swing.SwingUtilities;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.moduli.exceptions.ModuleNotFoundException;
import edu.dhbw.mannheim.tigers.moduli.listenerVariables.ModulesState;
import edu.dhbw.mannheim.tigers.sumatra.model.SumatraModel;
import edu.dhbw.mannheim.tigers.sumatra.model.data.airecord.IRecordFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.AIInfoFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.SimpleWorldFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.cam.CamDetectionFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.Agent;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.observer.IAIObserver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.observer.IWorldPredictorObserver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AAgent;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AReferee;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AWorldPredictor;
import edu.dhbw.mannheim.tigers.sumatra.presenter.moduli.IModuliStateObserver;
import edu.dhbw.mannheim.tigers.sumatra.presenter.moduli.ModuliStateAdapter;
import edu.dhbw.mannheim.tigers.sumatra.view.visualizer.VisualizerPanel;
import edu.dhbw.mannheim.tigers.sumatra.view.visualizer.internals.IFieldPanelObserver;
import edu.dhbw.mannheim.tigers.sumatra.view.visualizer.internals.IReplayOptionsPanelObserver;
import edu.dhbw.mannheim.tigers.sumatra.view.visualizer.internals.IRobotsPanelObserver;
import edu.dhbw.mannheim.tigers.sumatra.views.ISumatraView;
import edu.dhbw.mannheim.tigers.sumatra.views.ISumatraViewPresenter;


/**
 * Presenter for the visualizer.
 * <p>
 * NOTE: The fact that the view stores the actual state is not MVP-conform, but that would need greater refactoring and
 * I don't have time for this now
 * </p>
 * 
 * @author Bernhard, (Gero)
 */
public class VisualizerPresenter implements ISumatraViewPresenter, IRobotsPanelObserver, IFieldPanelObserver,
		IModuliStateObserver, IReplayOptionsPanelObserver
{
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	// Logger
	private static final Logger				log					= Logger.getLogger(VisualizerPresenter.class
																						.getName());
	
	private final VisualizerPanel				panel;
	private BotID									selectedRobotId	= BotID.createBotId();
	
	private final AgentVisualizerListener	agentListener		= new AgentVisualizerListener();
	private final WPVisualizerListener		wpListener			= new WPVisualizerListener();
	
	private final OptionsPanelPresenter		optionsPanelPresenter;
	
	private ScheduledExecutorService			execService;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * 
	 */
	public VisualizerPresenter()
	{
		panel = new VisualizerPanel();
		optionsPanelPresenter = new OptionsPanelPresenter(panel.getFieldPanel(), panel.getOptionsPanel());
		
		// --- register on fieldpanel as observer ---
		panel.getFieldPanel().addObserver(this);
		
		// --- register on optionspanel as observer ---
		panel.getOptionsPanel().addObserver(optionsPanelPresenter);
		
		// --- register on replay options panel
		panel.getReplayOptionsPanel().addObserver(this);
		
		// --- register on moduli ---
		ModuliStateAdapter.getInstance().addObserver(this);
	}
	
	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	
	// --------------------------------------------------------------------------
	// --- observer-methods -----------------------------------------------------
	// --------------------------------------------------------------------------
	@Override
	public void onRobotClick(final BotID botId)
	{
		// --- select/deselect item ---
		if (selectedRobotId.equals(botId))
		{
			selectedRobotId = BotID.createBotId();
			
		} else
		{
			selectedRobotId = botId;
		}
		
		panel.repaint();
		
	}
	
	
	@Override
	public void onFieldClick(final IVector2 posIn, final MouseEvent e)
	{
		boolean rightClick = SwingUtilities.isRightMouseButton(e);
		
		if (rightClick)
		{
			AReferee referee;
			try
			{
				referee = (AReferee) SumatraModel.getInstance().getModule(AReferee.MODULE_ID);
				referee.replaceBall(posIn);
			} catch (ModuleNotFoundException err)
			{
				log.error("Referee module not found.", err);
			}
			return;
		}
		
	}
	
	
	@Override
	public void onModuliStateChanged(final ModulesState state)
	{
		switch (state)
		{
			case ACTIVE:
				
				panel.getOptionsPanel().setInitialButtonState();
				panel.getOptionsPanel().setButtonsEnabled(true);
				panel.getFieldPanel().setPanelVisible(true);
				
				try
				{
					final SumatraModel model = SumatraModel.getInstance();
					
					Agent agent = (Agent) model.getModule(AAgent.MODULE_ID);
					agent.addObserver(agentListener);
					
					AWorldPredictor worldPredictor = (AWorldPredictor) model.getModule(AWorldPredictor.MODULE_ID);
					worldPredictor.addObserver(wpListener);
				} catch (final ModuleNotFoundException err)
				{
					log.error("no worldpredictor or botManager found!!!", err);
				}
				
				for (JCheckBox cb : panel.getOptionsPanel().getCheckBoxes().values())
				{
					optionsPanelPresenter.reactOnActionCommand(cb.getActionCommand(), cb.isSelected());
				}
				
				break;
			case NOT_LOADED:
				break;
			case RESOLVED:
				// --- clear connection-arrays ---
				if (execService != null)
				{
					execService.shutdown();
					try
					{
						execService.awaitTermination(2, TimeUnit.SECONDS);
					} catch (InterruptedException err)
					{
						log.error("Error waiting for termination.", err);
					}
				}
				panel.getOptionsPanel().setInitialButtonState();
				panel.getFieldPanel().setPanelVisible(false);
				panel.getFieldPanel().clearField();
				break;
			default:
				break;
		
		}
	}
	
	// --------------------------------------------------------------------------
	// --- inner-class-listener -------------------------------------------------
	// --------------------------------------------------------------------------
	
	protected class AgentVisualizerListener implements IAIObserver
	{
		
		
		@Override
		public void onNewAIInfoFrame(final AIInfoFrame lastAIInfoframe)
		{
			panel.getFieldPanel().updateAiFrame(lastAIInfoframe);
		}
		
		
		@Override
		public void onAIException(final Exception ex, final IRecordFrame frame, final IRecordFrame prevFrame)
		{
		}
	}
	
	
	protected class WPVisualizerListener implements IWorldPredictorObserver
	{
		private SimpleWorldFrame	lastWorldFrame	= null;
		
		
		@Override
		public void onNewWorldFrame(final SimpleWorldFrame wf)
		{
			setLastWorldFrame(wf);
			panel.getFieldPanel().updateWFrame(wf);
		}
		
		
		@Override
		public void onVisionSignalLost(final SimpleWorldFrame emptyWf)
		{
			setLastWorldFrame(emptyWf);
			panel.getFieldPanel().updateWFrame(emptyWf);
		}
		
		
		@Override
		public void onNewCamDetectionFrame(final CamDetectionFrame frame)
		{
		}
		
		
		/**
		 * @return the lastWorldFrame
		 */
		public SimpleWorldFrame getLastWorldFrame()
		{
			return lastWorldFrame;
		}
		
		
		/**
		 * @param lastWorldFrame the lastWorldFrame to set
		 */
		public void setLastWorldFrame(final SimpleWorldFrame lastWorldFrame)
		{
			this.lastWorldFrame = lastWorldFrame;
		}
	}
	
	
	// --------------------------------------------------------------------------
	// --- observer-methods -----------------------------------------------------
	// --------------------------------------------------------------------------
	
	
	@Override
	public void onRecord(final boolean active)
	{
		panel.getFieldPanel().getMultiLayer().setRecording(active, false);
		panel.getReplayLoadPanel().doUpdate();
	}
	
	
	@Override
	public void onSave(final boolean active)
	{
		panel.getFieldPanel().getMultiLayer().setRecording(active, true);
		panel.getReplayLoadPanel().doUpdate();
		
	}
	
	
	@Override
	public Component getComponent()
	{
		return panel;
	}
	
	
	@Override
	public ISumatraView getSumatraView()
	{
		return panel;
	}
	
	
	@Override
	public void onEmergencyStop()
	{
	}
	
}
