package edu.dhbw.mannheim.tigers.sumatra.model.data;

/*
 * *********************************************************
 * Copyright (c) 2009 - 2013, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Feb 1, 2013
 * Author(s): Daniel Andres <andreslopez.daniel@gmail.com>
 * 
 * *********************************************************
 */

import java.util.HashMap;
import java.util.Map;

import com.google.protobuf.Message;

import edu.dhbw.mannheim.tigers.sumatra.proto.AiControlAvailableProtos.AiControlAvailable;
import edu.dhbw.mannheim.tigers.sumatra.proto.AiControlProtos.AiControl;
import edu.dhbw.mannheim.tigers.sumatra.proto.AiControlStateProtos.AiControlState;
import edu.dhbw.mannheim.tigers.sumatra.proto.BotActionCommandProtos.BotActionCommand;
import edu.dhbw.mannheim.tigers.sumatra.proto.BotStatusProtos.BotStati;
import edu.dhbw.mannheim.tigers.sumatra.proto.LifeSignProtos.LifeSign;
import edu.dhbw.mannheim.tigers.sumatra.proto.LogMessagesProtos.LogMessage;
import edu.dhbw.mannheim.tigers.sumatra.proto.RefereeMsgProtos.RefereeCommandSimple;
import edu.dhbw.mannheim.tigers.sumatra.proto.model.RefereeProtos.RefereeMsg;
import edu.dhbw.mannheim.tigers.sumatra.proto.presenter.MainPresenterProtos.Emergency;
import edu.dhbw.mannheim.tigers.sumatra.proto.presenter.MainPresenterProtos.FilenameArray;
import edu.dhbw.mannheim.tigers.sumatra.proto.presenter.MainPresenterProtos.ModulesState;
import edu.dhbw.mannheim.tigers.sumatra.proto.presenter.TimerProtos.TimerInfo;
import edu.dhbw.mannheim.tigers.sumatra.util.log.LogProtos.FileTree;
import edu.dhbw.mannheim.tigers.sumatra.util.log.LogProtos.LogLevel;
import edu.dhbw.mannheim.tigers.sumatra.util.log.LogProtos.LoggingEventProto;
import edu.dhbw.mannheim.tigers.sumatra.util.messaging.MessageType;
import edu.dhbw.mannheim.tigers.sumatra.util.messaging.test.TestProtos;
import edu.dhbw.mannheim.tigers.sumatra.util.messaging.topics.ITopics;


/**
 * This enum contains all used topics and their matching protobuf type
 * @author Daniel Andres <andreslopez.daniel@gmail.com>
 */
public enum ETopics implements ITopics
{
	
	/**  */
	AI_CONTROL_AVAILABLE("android/aicontrolavail", MessageType.CONTROL, AiControlAvailable.getDefaultInstance()),
	/**  */
	AI_CONTROL("android/aicontrol", MessageType.NONDURABLE_INFO, AiControl.getDefaultInstance()),
	/**  */
	BOT_ACTION_COMMAND("android/botactioncmd", MessageType.NONDURABLE_INFO, BotActionCommand.getDefaultInstance()),
	/**  */
	BOT_STATI("android/botstati", MessageType.CONTROL, BotStati.getDefaultInstance()),
	/**  */
	LIFE_SIGN("android/lifesign", MessageType.NONDURABLE_INFO, LifeSign.getDefaultInstance()),
	/**  */
	LOG_MESSAGES("android/logmessages", MessageType.NONDURABLE_INFO, LogMessage.getDefaultInstance()),
	/**  */
	AI_CONTROL_STATE("android/emergencystop", MessageType.NONDURABLE_INFO, AiControlState.getDefaultInstance()),
	/**  */
	REFEREE_COMMAND("android/refereeCmd", MessageType.CONTROL, RefereeCommandSimple.getDefaultInstance()),
	
	/** Moduli State topic */
	MODULI_STATE("moduli/state", MessageType.CONTROL, ModulesState.getDefaultInstance()),
	/** */
	MODULI_CONFIG("moduli/config", MessageType.CONTROL, FilenameArray.getDefaultInstance()),
	
	/** Log record for fatal logs */
	LOG_FATAL("log/fatal", MessageType.NONDURABLE_INFO, LoggingEventProto.getDefaultInstance()),
	/** Log record for error logs */
	LOG_ERROR("log/error", MessageType.NONDURABLE_INFO, LoggingEventProto.getDefaultInstance()),
	/** Log record for warn logs */
	LOG_WARN("log/warn", MessageType.NONDURABLE_INFO, LoggingEventProto.getDefaultInstance()),
	/** Log record for info logs */
	LOG_INFO("log/info", MessageType.NONDURABLE_INFO, LoggingEventProto.getDefaultInstance()),
	/** Log record for debug logs */
	LOG_DEBUG("log/debug", MessageType.NONDURABLE_INFO, LoggingEventProto.getDefaultInstance()),
	/** Log record for trace logs */
	LOG_TRACE("log/trace", MessageType.NONDURABLE_INFO, LoggingEventProto.getDefaultInstance()),
	/** Log level of Sumatra */
	LOG_LEVEL("log/level", MessageType.CONTROL, LogLevel.getDefaultInstance()),
	/** File Tree for selecting filters in Log Panel */
	LOG_FILETREE("log/filetree", MessageType.CONTROL, FileTree.getDefaultInstance()),
	
	/** Referee commands: refMsg.proto */
	REFEREE_MSG("ref", MessageType.DURABLE_INFO, RefereeMsg.getDefaultInstance()),
	/** Referee commands that a user creates for its own: refMsg.proto */
	REFEREE_OWN("ref/own", MessageType.DURABLE_INFO, RefereeMsg.getDefaultInstance()),
	
	/** Sends timer info timer.proto */
	TIMER_INFO("timer", MessageType.NONDURABLE_INFO, TimerInfo.getDefaultInstance()),
	
	/** Sends emergency info */
	EMERGENCY("emergency", MessageType.CONTROL, Emergency.getDefaultInstance()),
	
	/** Test topic for local latency test */
	TEST_LATENCY_LOCAL("test/latency/local", MessageType.NONDURABLE_INFO, TestProtos.TimerInfo.getDefaultInstance()),
	/** Test topic for remote latency test - send (first step) */
	TEST_LATENCY_REMOTE_SND("test/latency/remote/send", MessageType.NONDURABLE_INFO, TestProtos.TimerInfo
			.getDefaultInstance()),
	/** Test topic for remote latency test - send (first step) */
	TEST_LATENCY_REMOTE_RCV("test/latency/remote/receive", MessageType.NONDURABLE_INFO, TestProtos.TimerInfo
			.getDefaultInstance()),
	/** Test topic for reconnect Test */
	TEST_RECONNECT_PUBLISH("test/connection/reconnect/publish", MessageType.NONDURABLE_INFO, TestProtos.ConnectionInfo
			.getDefaultInstance());
	
	/** Topic name of the MQTT Topic */
	private final String						mqttTopic;
	/** Control or Info type of expected message */
	private final MessageType				type;
	/** Wrapping type of expected message */
	private final Message					protoType;
	
	private static Map<String, ETopics>	nameValueMap	= new HashMap<String, ETopics>();
	
	static
	{
		for (final ETopics topic : ETopics.values())
		{
			nameValueMap.put(topic.getName(), topic);
		}
	}
	
	
	private ETopics(final String mqttTopic, final MessageType type, final Message protoType)
	{
		this.mqttTopic = mqttTopic;
		this.type = type;
		this.protoType = protoType;
	}
	
	
	/**
	 * Used to get the topic name
	 * @return the mqttTopic
	 */
	@Override
	public String getName()
	{
		return mqttTopic;
	}
	
	
	/**
	 * @return the type
	 */
	@Override
	public MessageType getType()
	{
		return type;
	}
	
	
	/**
	 * @return the protoType
	 */
	@Override
	public Message.Builder getProtoType()
	{
		return protoType.newBuilderForType();
	}
	
	
	/**
	 * @param name
	 * @return
	 */
	public static ETopics getTopicByName(final String name)
	{
		return nameValueMap.get(name);
	}
}
